#include <string.h>
#include <ctype.h>
#include <stdint.h>
#include "types.h"
#include "fat32.h"
#include "sd.h"

char sector_buffer[SECTOR_SIZE];
char fat_cache[SECTOR_SIZE];	        //this keeps us from reading the fat everytime we read a new sector
u32 current_fat_sector;
u32 current_data_sector;
u16 cur_dir_entry = 0;	                // number of files entrys in directory for specific extention


int strncasecmp(char *s1, char *s2, u16 n)
{
  if (n == 0)
    return 0;

  while (n-- != 0 && tolower(*s1) == tolower(*s2))
    {
      if (n == 0 || *s1 == '\0' || *s2 == '\0')
	break;
      s1++;
      s2++;
    }

  return tolower(*(unsigned char *) s1) - tolower(*(unsigned char *) s2);
}


//*****************************************************************************
// Parameter:  Sektor der gelesen werden soll
//             Datenpuffer
// Rckgabe:   Anzahl der Bytes die gelesen wurden
//
// Liest ein Daten- oder Fat-Sektor
//*****************************************************************************
u16 read_sector (const u08 fat, const u32 sector, const u16 num_of_bytes, char * data_buffer)
{
    if (fat) 
    {
        // Sektor wird nur gelesen, wenn er nicht gerade eben schon gelesen wurde
        if (current_fat_sector == sector) 
            return num_of_bytes;
        else 
        {
            current_fat_sector = sector;
            return card_read_sector(sector, num_of_bytes, data_buffer);
        }
    }
    else {
        // Sektor wird nur gelesen, wenn er nicht gerade eben schon gelesen wurde
        if (current_data_sector == sector)
            return num_of_bytes;
        else {
            current_data_sector = sector;
            return card_read_sector(sector, num_of_bytes, data_buffer);
        }
    }
}

//*****************************************************************************
// Rckgabe: Fehlernummer
//
// Initialisiert die Speicherkarte und liet Partionsinformationen aus dem MBR.
// Gibt den Status zurck.
//*****************************************************************************
u08 fat_init(void)
{
    u32 part_start;	
    u32 sectors_per_fat;
    u08 num_of_fats;   
    u16 RootDirSectors;
    u32 FATSz;
    u32 TotSec;
    u32 DataSec;
    u32 CountofClusters;
    
    
    if (card_detected() != 0) return E_ERROR;               // Wenn keine Speicherkarte vorhanden -> Fehler

    current_fat_sector = -1;                                // aktueller Fat- und Data-Sektor auf -1 (Init)
    current_data_sector = -1;

    card_init ();		                                    // Speicherkarte initialisieren

    read_sector (DATA, 0, SECTOR_SIZE, sector_buffer);      // gesamten Sektor (512 Bytes) mit Nummer 0 lesen und in im Puffer (sector_buffer) speichern     

    part_start = *((u32 *) & sector_buffer[454]);	        // Start der eigentlichen Partition
   
    read_sector (DATA, part_start, SECTOR_SIZE, sector_buffer);     // Den ersten Sektor aus der Partition auslesen (512 Bytes)

    // Erfassung des Fat Systems laut Datenblatt (Seite 14+)
    RootDirSectors = (( *((u16 *) & sector_buffer[17]) * 32) + ( *((u16 *) & sector_buffer[11]) - 1)) /  *((u16 *) & sector_buffer[11]);
    
    if (*((u16 *) & sector_buffer[22]) != 0)
        FATSz = *((u16 *) & sector_buffer[22]);
    else
        FATSz = *((u32 *) & sector_buffer[36]);
           
    if (*((u16 *) & sector_buffer[19]) != 0)
        TotSec = *((u16 *) & sector_buffer[19]);
    else
        TotSec = *((u32 *) & sector_buffer[32]);
    
    DataSec = TotSec - (*((u16 *) & sector_buffer[14]) + (*((u08 *) & sector_buffer[16]) * FATSz) + RootDirSectors);
    CountofClusters = DataSec / *((u08 *) & sector_buffer[13]);
   
    if (CountofClusters < 4085) 
    {
        printf("\r\nFehler: FAT12 gefunden (Fat32 erwartet)!\r\n");
        return E_INVALID;
    } 
    else if(CountofClusters < 65525) 
    {
        printf("\r\nFehler: FAT16 gefunden (Fat32 erwartet)!\r\n");
        return E_INVALID;
    } 
    else 
    {
        printf("\r\nFAT32 gefunden!\r\n");
    }
    
    // Partionsinformationen laut Datenblatt (Seite 9+) in Struct speichern
    partition.bytes_per_sector = *((u16 *) & sector_buffer[0x0b]);	    // Bytes pro Sektor
    partition.sectors_per_cluster = *((u08 *) & sector_buffer[0x0d]);	// Sektoren pro Cluster
    partition.reserved_sectors = *((u16 *) & sector_buffer[0x0e]);	    // Anzahl der reservierten Sektoren
    partition.reserved_sectors += part_start;                           // 
    
    num_of_fats = *((u08 *) & sector_buffer[0x10]);	                    // Anzahl der FATs
    sectors_per_fat = *((u32 *) & sector_buffer[0x24]);                 // Anzahl der Sektoren pro FAT

        // store Root Directorys first cluster, 32bit
    partition.root_directory = *((u32 *) & sector_buffer[0x2c]);	    // Erster Cluster in der Verzeichniskette
    partition.root_directory &= 0xfffffff;                              // Fat32 -> 28 bit; Bit 28 bis 31 auf Null
    partition.current_directory = partition.root_directory;             // aktuelles Verzeichnis Cluster auf Wurzel-Cluster setzen
    partition.data_starts = partition.reserved_sectors + sectors_per_fat * num_of_fats;     // Erster LBA (Logical Block Addressing) Cluster

    return 0;
}


//*****************************************************************************
// Parameter: Clusternummer
// Rckgabe:  Sektornumber
//
// Wandel Clusternummer in Sektornummer um
//*****************************************************************************
u32 cluster2sector (u32 cluster)
{
    // Berechnung laut Datenblatt (Seite 14)
    return partition.data_starts + ((cluster - 2) * partition.sectors_per_cluster);
}


//*****************************************************************************
// Parameter: Clusternummer
// Rckgabe:  Nchste Clusternummer
//
// Gibt die nchste Clusternummer in der Kette zurck
//*****************************************************************************
u32 get_next_cluster (u32 cluster)
{
    u32 sector_offset;
    u32 byte_offset;

    cluster &= 0xfffffff;	                                        //Fat32 -> 28 bit; die Bit 28 bis 31 auf Null
   
    //Berechnung des ClusterOffsets (Datenblatt Seite 15)
    sector_offset = cluster >> 7;                                                                    
    sector_offset += partition.reserved_sectors;                    // Offset = Cluster / 128 + Reservierte Sektoren
                                                                    // wie im Datenblatt: (N ist cluster,BPB_BytsPerSec = 512, ThisFATSecNum ist sector_offset)   
                                                                    // FATOffset = N * 4;
                                                                    // ThisFATSecNum = BPB_ResvdSecCnt + (FATOffset / BPB_BytsPerSec);
    
    if (sector_offset != current_fat_sector)                        // Der Sektor wird gelesen und in fat_cache gespeichert
    {
        read_sector(FAT, sector_offset, SECTOR_SIZE, fat_cache);
        current_fat_sector = sector_offset;
    }

    // Berechnung des ByteOffsets (Datenblatt Seite 15) 
    byte_offset = (cluster & 0x7f) << 2;                            // Gibt den Rest der Division von Cluster/512 zurck und multipoliziert diesen mit 4
                                                                    // wie im Datenblatt:
                                                                    // FATOffset = N * 4;
                                                                    // ThisFATEntOffset = REM(FATOffset / BPB_BytsPerSec);   

    byte_offset &= 0x1ff;                                           // Maximal 511 Bytes als Offset

    cluster = *((u32 *) & fat_cache[byte_offset]);                  // die Clusternummer wird aus dem fat_cache gelesen

    if ((cluster & 0xffffff8) == 0xffffff8)	return 0;               // Wenn Clusternummer grer als 0xffffff7, dann ist das Ende des Datentrgers erreicht

    return cluster & 0xfffffff;                                     // Clusternummer (28 bit) zurckgeben
}



//*****************************************************************************
// Parameter:  Dateiname
//             DateiInfo Struktur
// Rckgabe:   Fehlercode
//
// ffnet eine Datei
//*****************************************************************************

u08 fat_open (char * fname, FAT_FILE * file)
{
    char *nameptr;
    u32 sector = 0;
    u32 dir_cluster = 0;
    u08 sector_entry = 16; 
    u08 sector_count = 0;
    u08 seqnum = 0;		                // long name seqence number
    char localname[MAX_PATH] = { 0 };	// clear long name
    u08 m,n;
    
    if (card_detected() != 0) return E_ERROR;                               // Wenn Speicherkarte nicht vorhanden, Abbruch

    cur_dir_entry = 0;
    nameptr = localname;

    do {                                                                    // Datei suchen

        if (sector_entry >= 16) 
        {
            if (dir_cluster == 0) 
            {	                                                            // ersten VerzeichnisSektor lesen
                dir_cluster = partition.current_directory;	                // ersten Cluster vom Wurzelverzeichnis holen
                sector = cluster2sector(dir_cluster);	                    // Cluster in Sektor umwandeln
                sector_count = 0;
                read_sector(DATA, sector++, SECTOR_SIZE, sector_buffer);	// ersten Sektor vom Wurzelverzeichnis in den Puffer laden
            }
            else                                                            // nachdem der erste Sektor gelesen wurde, die nchsten lesen.. 
            {     
                if (++sector_count >= partition.sectors_per_cluster)        // wenn alle Sektoren im Cluster gelesen wurden 
                {
                    sector_count = 0;                                       // Sektorzhler auf Null
                    dir_cluster = get_next_cluster(dir_cluster);	        // nchsten Cluster berechnen
            
                    if (dir_cluster == 0) break;	                        // wenn der letzte Cluster im Verzeichnis erreicht ist, Abbruch
            
                    sector = cluster2sector (dir_cluster);	                // Cluster in Sektor umwandeln
                }
                
                read_sector (DATA, sector++, SECTOR_SIZE, sector_buffer);	// Nchsten Sektor im Wurzelverzeichnis lesen
            }
      
            sector_entry = 0;
            nameptr = sector_buffer;	                                    // Zeiger auf die Daten fr das Verzeichnis
        }
        // Alle Codes -> (Datenblatt Seite 23)
        if (*nameptr != 0xE5)	                                            // Wenn der Eintrag vorhanden ist, also nicht frei..
        {
            
            if (*(nameptr + 11) == ATTR_LFN)	                            // Attribute auslesen, checken ob es ein langer Name ist
            {			                                                    // Wenn ja, dann
                seqnum = (*nameptr & 0x3F) - 1;	                            // Sequenznummer (Datenblatt Seite 27) fr lange Namen ermitteln
                                                                            // nur die Bits 0 bis 5 sind relevant (Maskierung mit 0x3F)
                                                                            // Langen Namen erstellen laut Datenblatt
                localname[(u16)seqnum * 13 + 0] = *(nameptr + 1);
                localname[(u16)seqnum * 13 + 1] = *(nameptr + 3);
                localname[(u16)seqnum * 13 + 2] = *(nameptr + 5);
                localname[(u16)seqnum * 13 + 3] = *(nameptr + 7);
                localname[(u16)seqnum * 13 + 4] = *(nameptr + 9);
                localname[(u16)seqnum * 13 + 5] = *(nameptr + 14);
                localname[(u16)seqnum * 13 + 6] = *(nameptr + 16);
                localname[(u16)seqnum * 13 + 7] = *(nameptr + 18);
                localname[(u16)seqnum * 13 + 8] = *(nameptr + 20);
                localname[(u16)seqnum * 13 + 9] = *(nameptr + 22);
                localname[(u16)seqnum * 13 + 10] = *(nameptr + 24);
                localname[(u16)seqnum * 13 + 11] = *(nameptr + 28);
                localname[(u16)seqnum * 13 + 12] = *(nameptr + 30);
               
                if (*nameptr & 0x40) localname[(u16)seqnum * 13 + 13] = 0;  // Wenn nameptr mit 0x40 maskiert ist, dann war das die letzte Sequenz
                                                                            // der String wird mit 0 abgeschlossen
            }
            // Ansonsten handelt es sich um ein kurzen Namen (Datenblatt Seite 23 (Byte mit Offset 0 bis 10 beinhaltet den Name)
            else if ((*nameptr != '.') && ((*(nameptr + 11) & (ATTR_HIDDEN | ATTR_SYSTEM | ATTR_VOLUME))== 0)) 
            {           
                if (localname[0] == 0) 
                {
                    // Zeichen kopieren laut Datenblatt
                    localname[0] = *(nameptr);                              // 8 Bytes (mit Offset 0 bis 7) ist der Dateiname 
                    localname[1] = *(nameptr + 1);
                    localname[2] = *(nameptr + 2);
                    localname[3] = *(nameptr + 3);
                    localname[4] = *(nameptr + 4);
                    localname[5] = *(nameptr + 5);
                    localname[6] = *(nameptr + 6);
                    localname[7] = *(nameptr + 7);
                    localname[8] = '.';                                     // Punkt setzen
                    localname[9] = *(nameptr + 8);                          // 3 Byte (mit Offset 8 bis 10) ist die Dateiendung
                    localname[10] = *(nameptr + 9);
                    localname[11] = *(nameptr + 10);
                    localname[12] = 0;	                                    // String abschlieen
            
                    
                    // "Space" - Zeichen entfernen
                    for (m = 0; m < 7; m++)                                 // An welcher Stelle (m) ist ein Space Zeichen
                        if (localname[m] == ' ') break;
                    for (n=0; n < 5; n++)                                   // Ab der Stelle m die Dateiendung auffllen
                        localname[(u16)m+n] = localname[(u16)8 + n];
                }
                            
                // berprfen ob gesuchter und gefundener Dateiname bereinstimmen
                if (strncasecmp(localname, fname, MAX_PATH) == 0) 
                {
                    // Datei Informationen fllen (Datenblatt Seite 23)
                    file->first_cluster = *((u16 *) & nameptr[0x14]);
                    file->first_cluster <<= 16;
                    file->first_cluster += *((u16 *) & nameptr[0x1a]);
                    file->first_cluster &= 0xfffffff;
                    file->file_size = *((u32 *) & nameptr[0x1c]);
                    file->current_cluster = file->first_cluster;
                    file->pointer = 0;
                    
                    return 0;                                               // Datei gefunden, Abbruch
                }
                        
                localname[0] = 0;	                                        // Langen Namen Lschen
            }   
        }	          

        nameptr += 32;                                                      
        sector_entry++;
    } while (1);

    return E_INVALID;		                                                // Fehler, Datei nicht gefunden
}



//*****************************************************************************
// Parameter: Datenpuffer in dem die Daten gespeichert werden
//            Anzahl der Bytes die gelesen werden sollen
//            Datei Struktur
// Rckgabe:  Anzahl der gelesenen Bytes
//
// Liest eine bestimmte Anzahl an Bytes einer Datei
//*****************************************************************************
u16 fat_read (char * out, u16 size, FAT_FILE * file)
{
    u16 read = 0;		// number of bytes read
    char *ptr;
    
    if (card_detected() != 0) return 0;                         // Wenn Speicherkarte nicht vorhanden, Abbruch
    if (!size) return size;                                     // Wenn angeforderte Gre 0, dann Abbruch
    if (size > SECTOR_SIZE) size = SECTOR_SIZE;                 // Wenn angeforderte Gre grer der Sektorgre, size auf Sektorgre

    if (file->pointer + size > file->file_size)                 // Wenn die angeforderte Gre der Gre der Datei berschreitet bzw. auerhalb der Datei liegt,
        size = (u16) (file->file_size - file->pointer);         // dann werden die restlichen Bytes der Datei zurckgegeben


    do {
        // Sektoroffset berechnen, in welchem Sektor befinden wir uns gerade
        u08 sector_in_cluster = (u08) ((file->pointer / partition.bytes_per_sector) % partition.sectors_per_cluster);
      
        ptr = &sector_buffer[file->pointer % partition.bytes_per_sector];   
      
        read_sector (DATA,cluster2sector (file->current_cluster) + sector_in_cluster, SECTOR_SIZE, sector_buffer);
      
        for (; size > 0; size--) 
        {
            if (ptr >= (sector_buffer + sizeof (sector_buffer))) break;             
            *out++ = *ptr++;
            file->pointer++;
            read++;
        }
      
        // berprfen ob das Ende eines Clusters erreicht ist
        if ((sector_in_cluster + 1) % partition.sectors_per_cluster == 0)
            file->current_cluster = get_next_cluster (file->current_cluster);   // nchsten Cluster laden
      
    } while (size);                                                             // solange bis alle angeforderteten Bytes gelesen wurden

    return read;		                                                        // Anzahl der gelesenen Bytes zurckgeben
}


//*****************************************************************************
// Parameter: Datei die durchsucht werden soll
//            Offset innerhalb der Datei
//            Startpunkt
//
// Setzt den Dateizeiger auf die bergenene Position innerhalb der Datei
//*****************************************************************************
void fat_seek (FAT_FILE * file, u32 offset)
{
    u32 cluster_offset;
    
    file->pointer = offset;
    
    if (file->pointer > file->file_size) file->pointer = file->file_size;

    cluster_offset = file->pointer / ( (u32) partition.bytes_per_sector * partition.sectors_per_cluster);

    file->current_cluster = file->first_cluster;

    for (; cluster_offset > 0; --cluster_offset)
        file->current_cluster = get_next_cluster (file->current_cluster);
}


//*****************************************************************************
// Parameter: Dateistruktur die durchsucht werden soll
//            Offset innerhalb der Datei
//            Anzahl der Bytes die gelesen werden sollen
//            Puffer fr gelesende Daten
//
// Liest eine bestimmte Anzahl an Bytes von einer bestimmten Position innerhalb einer Datei
//*****************************************************************************

void fat_read_bytes(FAT_FILE * file, u32 offset, u16 size, char * out)   
{
        fat_seek (file, offset);
        fat_read (out, size, file);
}
