#include "types.h"
#include "sd.h"
#include "cpu.h"



//*****************************************************************************
// Parameter: DatenByte das versendet werden soll
// Rckgabe:  Empfangendes DatenByte
//
// Sendet ein Datenbyte bers SPI und empfngt ein Byte
//*****************************************************************************
u08 card_sendbyte(u08 data)
{
    u16 i;
    SPDR = data;
    i = 0x1000;
    while (!(SPSR & (1 << SPIF))) {     // warten bis Byte gesendet wurde
        if (i-- == 0) break;            // Abbruch nach 4096 durchlufen
    }
    return SPDR;
}


//*****************************************************************************
// Rckgabe: erfolgreich (0)
//           Fehler (!0)
//
// Speicherkarte initialisieren und den Status zurckgeben
//*****************************************************************************
u08 card_init(void)
{
    u08 result;		                // Antwort der Speicherkarte
    u16 retry;		                // Abbruchzhler
    
    CARD_PORT |= (1 << CARD_CS);	// CS Port auf High -> Speicherkarte deaktiviert
    CARD_DIR |=  (1 << CARD_CS);	// CS Port als Output

    /*
    for (retry = 0; retry < 11; retry++) 
    {
        card_sendbyte(0xFF);
    }
    */
    
    retry = 0;
    
    do 
    {
        result = card_sendcommand(CARD_GO_IDLE_STATE, 0);	    // Speicherkarte in den Idle-Modus versetzen
        if (retry++ == 100) break;                              // nach 100 versuchen abbrechen 
    } while (result != CARD_R1_IDLE_STATE);                     // Solange wiederholen bis IDLE-STATE Antwort zurckkommt

    
    if (result == 0) return -1;		                            // Wenn Antwort 0 -> Fehler
    
    retry = 0;
    
    do 
    {
        result = card_sendcommand(CARD_SEND_OP_COND, 0);	    // Speicherkarte aktivieren
        if (retry++ == 1000) break;                             // nach 1000 versuchen abbrechen
    } while (result);                                           // Wenn Antwort !0 -> Karte aktiv

    retry = 0;
    
    do 
    {
        result = card_sendcommand(CARD_CRC_ON_OFF, 0);	        // CRC ausschalten
        if (retry++ == 100) break;                              // nach 100 versuchen abbrechen
    } while (result);                                           // Wenn Antwort !0 -> Kommando erfolgreich

    retry = 0;
    
    do 
    {
        result = card_sendcommand(CARD_SET_BLOCKLEN, 512);	    // Blocklnge auf 512 Bytes setzen
        if (retry++ == 100) break;                              // nach 100 versuchen abbrechen
    } while (result);                                           // Wenn Antwort !0 -> Kommando erfolgreich

    return 0;
}


//*****************************************************************************
// Parameter: Befehls-Byte
//            Argument
// Rckgabe:  Empfangenes Antwortbyte
//
// Ein Befehl zur Speicherkarte senden (Aktivieren/Deaktivieren der Speicherkarte)
//*****************************************************************************
u08 card_sendcommand(u08 cmd, u32 arg)
{
    u08 resp;                       // Antwort der Speicherkarte

    CARD_SELECT	                    // Speicherkarte aktivieren

    resp = card_command(cmd, arg);	// Befehl senden

    CARD_DESELECT	                // Speicherkarte deaktivieren
    
    card_sendbyte(0xFF);	        // Dummybyte senden

    return resp;                    // Antwort zurckgeben
}


//*****************************************************************************
// Parameter: Befehls-Byte
//            Argument
// Rckgabe:  Empfangenes Antwortbyte
//
// Eigentliches Senden eines Befehls zur Speicherkarte
// (Ohne Aktivieren/Deaktivieren der Speicherkarte)
//*****************************************************************************
u08 card_command(u08 cmd, u32 arg)
{

    u08 resp;		                // Antwort der Speicherkarte
    u08 retry = 0;		            // Abbruchzhler

    card_sendbyte(cmd);		        // Befehl senden
    
    card_sendbyte(arg >> 24);	    // Argument senden (4 Bytes)
    card_sendbyte(arg >> 16);
    card_sendbyte(arg >> 8);
    card_sendbyte(arg);
    
    card_sendbyte(0x95);	        // "CRC Prfung erfolgreich" senden

    // Warten bis Karte wieder antwortet (Antwort != 0xFF)
    while ((resp = card_sendbyte(0xff)) == 0xff) 
    {
        if (retry++ > 8) break;
    }

    return resp;		            // Antwort zurckgeben
}


//*****************************************************************************
// Parameter: Nummer des Sektors in der Partition
//            Datenpuffer
// Rckgabe:  Anzahl der gelesenen Bytes
//
// Liest Daten aus einem Sektor.
// Die Sektornummern starten am Anfang der Partition und nicht im MBR.
// Sektornummer 0 liet nicht den MBR sondern die ID des Datentrgers.
//*****************************************************************************
u16 card_read_sector(u32 sec, u16 num_of_bytes, char * buffer)
{
    u16 i = 0;
    u16 count = 0;
    
    card_busy_wait ();		            // Warten bis die Speicherkarte bereit ist

    CARD_SELECT                         // Karte aktivieren

    if (0 == card_command(CARD_READ_SINGLE_BLOCK, sec << 9))    // Befehl zum lesen eines Sektors (sec) senden; sec = sec * 512 (da 512 Bytes pro Sektor) 
    {                                                           // Wenn die Antwort 0 ist
        if (card_response(0xFE))                                // Dann solange warten bis die Speicherkarte den Code 0xFE zurckgibt
        {
            for (i = 0; i < num_of_bytes; i++)                  // Anzahl der Bytes in den Puffer packen, die gelesen werden sollen
            {
                *(buffer + i) = card_sendbyte(0xFF);            
            }
            
            for (i = 0; i < SECTOR_SIZE - num_of_bytes; i++)    // Da immer der ganze Sektor bertragen wird, mssen noch die restlichen Bytes                                                                
            {                                                   // bertragen werden (werden aber ignoriert)
                card_sendbyte(0xFF);
            }
        }
    }

    CARD_DESELECT                       // Karte deaktivieren

    card_sendbyte(0xff);                // Dummybyte senden

    return count;                       // zum vergleich ob auch alle Bytes gelesen wurden, wird die Anzahl der gelesenen Bytes zurckgegeben
}


//*****************************************************************************
// Rckgabe: bereit (0)
//           beschftigt (1)
//
// Warten bis die Speicherkarte bereit ist
//*****************************************************************************
u08 card_busy_wait(void)
{
    u16 i;

    CARD_DESELECT           // Speicherkarte deaktivieren 
    card_sendbyte(0xff);	// 8 Taktzyklen warten
    CARD_SELECT             // Speicherkarte aktivieren
    card_sendbyte (0xff);	// 32 Taktzyklen warten
    card_sendbyte (0xff);	
    card_sendbyte (0xff);	
    card_sendbyte (0xff);	

    i = 25000;

    while (i--) 
    {
        if (card_sendbyte(0xff) == 0xff) break;		// Warten bis Speicherkarte bereit ist (0xFF Rckgabe)
    }
    
    if (i == 0) return 1;		                    // Abbruch nach 25000 mal, Karte ist nicht bereit

    return 0;			                            // Karte ist bereit
}

//*****************************************************************************
// Parameter: Erwartete Antwort
// Rckgabe: richtige Antwort (1)
//           falsche Antwort (0)
//
// Warten bis die Speicherkarte die erwartete Antwort zurckgibt
//*****************************************************************************
u08 card_response(u08 res)
{
    u16 count = 50000;
    while (count > 0) 
    {
        if (card_sendbyte (0xFF) == res) return 1;		// Wenn erwartete Antwort von der Karte zurckkommt, Abbruch
        count--;
    }
    return 0;                                           // Wenn nach 50000 mal nachfragen die falsche Antwort geliefert wird -> Fehler
}

//*****************************************************************************
// Rckgabe: Speicherkarte gefunden (0)
//           Speicherkarte nicht gefunden (1)
//
// Prfen ob eine Speicherkarte im Slot ist
//*****************************************************************************
u08 card_detected(void)
{
    CARD_PORT |= (1 << CARD_DET);                       // Detect-Pin pull-up aktivieren
    CARD_DIR &= ~(1 << CARD_DET);                       // Detect-Pin als Input
       
    if (!(CARD_INP & (1<<CARD_DET)))                    // Wenn Detect-Pin Low ist, dann Karte vorhanden  
        return 0;
        
    return 1;                                           // Karte nicht gefunden      
}
